package br.ufpr.inf.rp09;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintStream;
import java.net.MalformedURLException;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.util.Date;
import java.util.StringTokenizer;
import org.apache.jmeter.engine.ClientJMeterEngine;
import org.apache.jmeter.engine.StandardJMeterEngine;
import org.apache.jmeter.save.SaveService;
import org.apache.jmeter.util.JMeterUtils;
import br.ufpr.inf.rp09.controller.PostgreSQLTestController;
import br.ufpr.inf.rp09.controller.VoltDBTestController;
import br.ufpr.inf.rp09.exceptions.CommandLineArgumentException;
import br.ufpr.inf.rp09.exceptions.JMeterPropertiesException;
import br.ufpr.inf.rp09.testdefinitions.enumerations.TestCaseEnum;

/**
 * Classe principal ModastJMeter - processa as opções de entrada, 
 * inicializa valores de configuração, o controlador de testes
 * cria os clientes JMeter e chama a execução dos testes
 * 
 * @author Rodrigo Piassetta / GRR20094994
 */
public class ModastJMeter {

	/* constantes de configuracao */
	final static String JMETER_PATH = "jmeter";
	final static String JMETER_PROPERTIES = JMETER_PATH+"/bin/jmeter.properties";
	final static String OUTPUT_RESULTS_PATH = "logs/";
	final static String OUTPUT_RESULTS_PREFIX = "test_results_";
	final static String OUTPUT_RESULTS_EXT = ".log";
	final static String REMOTE_HOSTS_PROPERTY = "remote_hosts";
	final static String LOCAL_HOSTNAME = "localhost";
	final static String LOCAL_IPADDRESS = "127.0.0.1";
	/* numero obrigatorio de argumentos */ 
	private static final int NUM_ARGS = 16;
	/* prefixos dos argumentos de linha de comando */
	private static final String TESTCASE_ARGUMENT = "-tc";
	private static final String TEST_TYPE_ARGUMENT = "-tp";
	private static final String DATABASE_ADDRESS_ARGUMENT = "-ip";
	private static final String DATABASE_PORT_ARGUMENT = "-p";
	private static final String DATABASE_NAME_ARGUMENT = "-db";
	private static final String DATABASE_USERNAME_ARGUMENT = "-u";
	private static final String DATABASE_PASSWORD_ARGUMENT = "-pwd";
	private static final String SAVE_LOGFILES_ARGUMENT = "-savelog";
	private static final String TEST_STEPS_ARGUMENT = "-s";
	/* tipos de testes */
	private static final String CONNECTION_TESTCASE = "c";
	private static final String TRANSACTIONS_TESTCASE = "t";
	private static final String VOLTDB_TEST_TYPE = "volt";
	private static final String POSTGRESQL_TEST_TYPE = "post";
	/* mensagens de excecoes/erros  */
	private static final String USAGE_MSG = "Usage: java -jar ModastJmeter -tc testCase[c|t] -tp [volt|post] -ip dbIpAdress -p dbPort -db databaseName -u username -pwd password -s testStepsTotal";
	private static final String INVALID_ARG_MSG = "Invalid argument! "+USAGE_MSG;	
	private static final String LOAD_PROPERTIES_EXCEPTION_MSG = "Erro ao carregar o arquivo jmeter.properties";
	private static final String TEST_CLIENT_EXCEPTION_MSG = "Nao pode contactar o cliente. Cliente não sera adicionado ao teste.";
	private static final String EXECUTE_TEST_EXCEPTION_MSG = "Falha durante a execucao dos testes. Teste sera encerrado.";
	private static final String END_APP_MSG = "Testes finalizados. Encerrando aplicacao.";
	/* atributos de inicializacao dos testes */
	private static TestCaseEnum selectedTestCase; // caso de teste a ser executado
	private static String selectedTestType;
	private static String databaseIpAddress;
	private static String databasePort;
	private static String databaseName;
	private static String username;
	private static String password;
	private static PrintStream outputResults;
	private static Long testStepsTotal;
	//private static boolean saveLogFiles = false; sempre salvar os logs
	private static boolean saveLogFiles = true;	
	
    /**
     * Recebe os argumentos da linha de comando e define como os testes devem ser executados.
     * Inicia os valores de configuracao, o controlador de testes,
     * cria os clientes de teste e chama a execucao dos testes
     * 
     * (Usage: java -jar ModastJmeter -tc testCase[c|t] -tp [volt|post] -ip dbIpAdress -p dbPort -db databaseName -u username -pwd password -s testStepsTotal).
     * 
     * @param args Os argumentos de execucao dos testes
     */	
	public static void main(String[] args) {

		try {
			//processa argumentos de linha de comando
			processInputArgs(args);
			
			//cria arquivo para impressao dos resultados dos testes
			File outputResultsFolder = new File(OUTPUT_RESULTS_PATH);
			if(!outputResultsFolder.exists())
				outputResultsFolder.mkdir();
			outputResults = new PrintStream(new File(OUTPUT_RESULTS_PATH+OUTPUT_RESULTS_PREFIX+System.currentTimeMillis()+OUTPUT_RESULTS_EXT));
			
			//inicializa configuracoes do JMeter
			initializeJMeterSettings();

			//cria controlador de testes de acordo com caso de teste e o tipo de teste escolhidos
			//(stress de conexoes / stress de transacoes)
			//(voltDB / PostgreSQL / HTTP Request)
			switch(selectedTestType) {
				case VOLTDB_TEST_TYPE:
					executeVoltDBTest();
					break;
				case POSTGRESQL_TEST_TYPE:
					executePostgreSQLTest();
					break;
			}
		} catch (CommandLineArgumentException e) {
			//excecao vinda de processInputArgs(args)
			System.err.println(e.getMessage());
			System.exit(1);
		} catch (FileNotFoundException e) {
			//excecao vinda de new PrintStream()
			System.err.println(e.getMessage());
			System.exit(1);
		} catch (JMeterPropertiesException e) {
			//excecao vinda de initializeJMeterSettings()
			System.err.println(e.getMessage());
			System.exit(1);
		}

		System.out.println(END_APP_MSG+" "+new Date(System.currentTimeMillis()));
		System.exit(0);
	}

	/**
	 * Processa os argumentos da linha de comando
	 *  
	 * @param args
	 * 
	 * @throws CommandLineArgumentException Se houver algum argumento invalido ou insuficiente
	 */
	private static void processInputArgs(String[] args) throws CommandLineArgumentException {
		int i = 0;
		String strArg;
		if(args.length < NUM_ARGS) {
			throw new CommandLineArgumentException(INVALID_ARG_MSG);	
		}

		while(i < args.length) {
			if(args[i].compareTo(TESTCASE_ARGUMENT) == 0) {
				 strArg = args[++i];
				 if(strArg.compareTo(CONNECTION_TESTCASE) == 0) {
					 selectedTestCase = TestCaseEnum.CONNECTIONS_STRESS_TESTCASE;
				 }
				 else if(strArg.compareTo(TRANSACTIONS_TESTCASE) == 0) {
					 selectedTestCase = TestCaseEnum.TRANSACTIONS_STRESS_TESTCASE;
				 }
				 else {
					 throw new CommandLineArgumentException(INVALID_ARG_MSG);
				 }
			}
			else if(args[i].compareTo(TEST_TYPE_ARGUMENT) == 0) {
				 strArg = args[++i];
				 if(strArg.compareTo(VOLTDB_TEST_TYPE) == 0
						 || strArg.compareTo(POSTGRESQL_TEST_TYPE) == 0) {
					 selectedTestType = strArg;
				 }
				 else {
					 throw new CommandLineArgumentException(INVALID_ARG_MSG);
				 }				
			}
			else if(args[i].compareTo(DATABASE_ADDRESS_ARGUMENT) == 0) {
				strArg = args[++i];
				if(strArg.length() > 0) {
					databaseIpAddress = strArg;
				} else {
					throw new CommandLineArgumentException(INVALID_ARG_MSG);
				}				
			}
			else if(args[i].compareTo(DATABASE_PORT_ARGUMENT) == 0) {
				strArg = args[++i];
				if(strArg.length() > 0) {
					databasePort = strArg;
				} else {
					throw new CommandLineArgumentException(INVALID_ARG_MSG);
				}				
			}
			else if(args[i].compareTo(DATABASE_NAME_ARGUMENT) == 0) {
				strArg = args[++i];
				if(strArg.length() > 0) {
					databaseName = strArg;
				} else {
					throw new CommandLineArgumentException(INVALID_ARG_MSG);
				}			
			}
			else if(args[i].compareTo(DATABASE_USERNAME_ARGUMENT) == 0) {
				strArg = args[++i];
				if(strArg.length() > 0) {
					username = strArg;
				} else {
					throw new CommandLineArgumentException(INVALID_ARG_MSG);
				}			
			}
			else if(args[i].compareTo(DATABASE_PASSWORD_ARGUMENT) == 0) {
				strArg = args[++i];
				if(strArg.length() > 0) {
					password = strArg;
				} else {
					throw new CommandLineArgumentException(INVALID_ARG_MSG);
				}
			}
			else if(args[i].compareTo(TEST_STEPS_ARGUMENT) == 0) {
				strArg = args[++i];
				if(strArg.length() > 0) {
					testStepsTotal = Long.valueOf(strArg);
				} else {
					throw new CommandLineArgumentException(INVALID_ARG_MSG);
				}				
			}
			else if(args[i].compareTo(SAVE_LOGFILES_ARGUMENT) == 0) {
				saveLogFiles = true;
			}
			
			i++;
		}		
	}

	/**
	 * Inicializa as configurações do arquivo "jmeter.properties" para execucao do JMeter.
	 * 
	 * @throws JMeterPropertiesException Se nao e possivel carregar as configuracoes do arquivo de propriedades do JMeter 
	 */
	private static void initializeJMeterSettings() throws JMeterPropertiesException {
		// Configuracao inicial
		JMeterUtils.setJMeterHome(JMETER_PATH);		
		JMeterUtils.loadJMeterProperties(JMETER_PROPERTIES);
		JMeterUtils.initLogging();
		JMeterUtils.initLocale();

		try {
			SaveService.loadProperties();
		} catch (IOException e) {
			throw new JMeterPropertiesException(LOAD_PROPERTIES_EXCEPTION_MSG, e.getCause());
		}		
	}
	
	private static void executeVoltDBTest() {
		VoltDBTestController testController = new VoltDBTestController(selectedTestCase);
		
		//inicializa valores de conexao com o banco de dados a ser testado
		testController.setDatabaseAddress(databaseIpAddress);
		testController.setDatabasePort(databasePort);
		testController.setDatabaseName(databaseName);
		testController.setUsername(username);
		testController.setPassword(password);		
		
		//criando clientes de teste
		String testClientAddress;
		String strRemoteHosts = JMeterUtils.getJMeterProperties().getProperty(REMOTE_HOSTS_PROPERTY);
		StringTokenizer tokenRemoteHosts = new StringTokenizer(strRemoteHosts, ",");
		while (tokenRemoteHosts.hasMoreElements()) {
			testClientAddress = (String)tokenRemoteHosts.nextElement();
			System.out.println("connecting to "+testClientAddress);
			if(testClientAddress.compareTo(LOCAL_HOSTNAME) == 0 || testClientAddress.compareTo(LOCAL_IPADDRESS) == 0) {
				// cria cliente de teste local / cliente principal
				StandardJMeterEngine testClient = new StandardJMeterEngine();
				testController.setMainTestClient(testClient);
			} else {
				try {
					// cria cliente de teste remoto					
					ClientJMeterEngine testClient = new ClientJMeterEngine(testClientAddress);
					testController.addRemoteTestClient(testClientAddress, testClient);
				} catch (RemoteException | MalformedURLException | NotBoundException e) {
					System.out.println("<"+testClientAddress+"> "+e.getCause());
					System.out.println(TEST_CLIENT_EXCEPTION_MSG);
				}
			}
		}
		
		//executa os testes ate que a maquina de estados atinja o estado final
		long i = 0;
		while(i < testStepsTotal && testController.getModastStateMachine().getState() != testController.getModastStateMachine().getFinalState()) {
			try {
				testController.executeTest();
			} catch (Exception e) {
				System.out.println(e.getMessage()+" - "+e.getCause());
				System.out.println(EXECUTE_TEST_EXCEPTION_MSG);
			}
			//imprime resultados dos testes
			testController.printResults(outputResults);
			i++;
		}
		
		//apaga ou arquiva os arquivos de log dos testes
		testController.removeLogFiles(saveLogFiles);
	}

	private static void executePostgreSQLTest() {
		PostgreSQLTestController testController = new PostgreSQLTestController(selectedTestCase);
		
		//inicializa valores de conexao com o banco de dados a ser testado
		testController.setDatabaseAddress(databaseIpAddress);
		testController.setDatabasePort(databasePort);
		testController.setDatabaseName(databaseName);
		testController.setUsername(username);
		testController.setPassword(password);		
		
		//criando clientes de teste
		String testClientAddress;
		String strRemoteHosts = JMeterUtils.getJMeterProperties().getProperty(REMOTE_HOSTS_PROPERTY);
		StringTokenizer tokenRemoteHosts = new StringTokenizer(strRemoteHosts, ",");
		while (tokenRemoteHosts.hasMoreElements()) {
			testClientAddress = (String)tokenRemoteHosts.nextElement();
			System.out.println("connecting to "+testClientAddress);
			if(testClientAddress.compareTo(LOCAL_HOSTNAME) == 0 || testClientAddress.compareTo(LOCAL_IPADDRESS) == 0) {
				// cria cliente de teste local / cliente principal
				StandardJMeterEngine testClient = new StandardJMeterEngine();
				testController.setMainTestClient(testClient);
			} else {
				try {
					// cria cliente de teste remoto					
					ClientJMeterEngine testClient = new ClientJMeterEngine(testClientAddress);
					testController.addRemoteTestClient(testClientAddress, testClient);
				} catch (RemoteException | MalformedURLException | NotBoundException e) {
					System.out.println("<"+testClientAddress+"> "+e.getCause());
					System.out.println(TEST_CLIENT_EXCEPTION_MSG);
				}
			}
		}
		
		//executa os testes ate que a maquina de estados atinja o estado final
		long i = 0;
		while(i < testStepsTotal && testController.getModastStateMachine().getState() != testController.getModastStateMachine().getFinalState()) {
			try {
				testController.executeTest();
			} catch (Exception e) {
				System.out.println(e.getMessage()+" - "+e.getCause());
				System.out.println(EXECUTE_TEST_EXCEPTION_MSG);
			}
			//imprime resultados dos testes
			testController.printResults(outputResults);
			i++;
		}
		
		//apaga ou arquiva os arquivos de log dos testes
		testController.removeLogFiles(saveLogFiles);
	}
}
