package br.ufpr.inf.rp09.controller;

/**
 * Classe abstrata do controlador de testes para testes com clientes JMeter.
 * O teste e integralmente centralizado nesta classe.
 * Esta classe e responsavel por:<ul>
 * <li>configurar o plano de teste,</li>
 * <li>aguardar pelo encerramento da iteracao de teste em todos os clientes</li>
 * <li>direcionar os resultados para avaliacao</li>
 * <li>verificar se houve transicao de estado na maquina de estados</li> 
 * </ul>
 * @author Rodrigo Piassetta / GRR20094994
 */
public abstract class AbstractTestController {

	/**
	 * Algoritmo obrigatorio para instanciamento de uma classe filha
	 */
	protected AbstractTestController() {
		//cria maquina de estados
		createStateMachine();
		//define step inicial
		initTestStep();
		//cria monitor de resultados
		createMonitorLogResults();
		//cria o objeto que agrupa os resultados
		createTestResultAggregator();
	}

	/**
	 * Cria a maquina de estados que controla o estado do sistema em teste
	 */
	protected abstract void createStateMachine();
	
	/**
	 * Configura para a iteracao de teste inicial
	 */
	protected abstract void initTestStep();
	
	/**
	 * Cria o monitor dos logs de resultados
	 */
	protected abstract void createMonitorLogResults();
	
	/**
	 * Cria o objeto responsvel por agrupar os resultados dos testes
	 */
	protected abstract void createTestResultAggregator();
	
	/**
	 * Rotina de execucao de uma iteracao de teste.
	 * 
	 * @throws Exception Se nao for possivel executar o plano de testes por qualquer motivo
	 */
	public void executeTest() throws Exception {
		//configura plano de teste
		configureTestPlan();

		//configura monitor do log de resultados
		configureMonitorLogResults();
		
		//timestamp de inicio do teste
		long testStartTime = System.currentTimeMillis();
		//executa clientes de teste
		runTestClients();
		while(isTestRunning()); //aguarda pelo encerramento dos testes
		//timestamp de encerramento do teste
		long testEndTime = System.currentTimeMillis();

		//processa resultados do teste
		processTestResults();
		
		//avalia e armazena resultados 
		evaluateTestResults(testStartTime, testEndTime);

		//verifica se ha transicao de estado		
		checkStateTransition();
		
		//configura para proxima iteracao
		nextTestStep();		
	}

	/**
	 * Configura o plano de testes
	 */
	protected abstract void configureTestPlan();
	
	/**
	 * Configura monitor do log de resultados
	 */
	protected abstract void configureMonitorLogResults();
	
	/**
	 * Executa o plano de testes nos clientes de teste 
	 * 
	 * @throws Exception Se nao for possivel executar o plano de testes por qualquer motivo 
	 */	
	protected abstract void runTestClients() throws Exception;
	
	/**
	 * Verifica se o teste esta em execucao
	 * 
	 * @return true se o plano de testes esta em execucao, senao false
	 */	
	protected abstract boolean isTestRunning();
	
	/**
	 * Processa os resultados dos testes
	 */
	protected abstract void processTestResults();
	
	/**
	 * Avalia os resultados do teste
	 * 
	 * @param testStartTime timestamp do inicio da iteracao
	 * @param testEndTime timestamp do final da iteracao
	 */
	protected abstract void evaluateTestResults(long testStartTime, long testEndTime);
	
	/**
	 * Verifica se houve transicao de estado
	 */		
	protected abstract void checkStateTransition();
	
	/**
	 * Configura o controlado de teste para a proxima iteracao
	 */
	protected abstract void nextTestStep();
}
