package br.ufpr.inf.rp09.controller;

import java.io.File;
import java.io.PrintStream;
import java.util.HashMap;
import java.util.Map;

import org.apache.jmeter.engine.ClientJMeterEngine;
import org.apache.jmeter.engine.JMeterEngineException;
import org.apache.jmeter.engine.StandardJMeterEngine;

import br.ufpr.inf.rp09.exceptions.TestClientException;
import br.ufpr.inf.rp09.statemachine.ModastStateEvaluator;
import br.ufpr.inf.rp09.statemachine.ModastStateMachine;
import br.ufpr.inf.rp09.statemachine.enumerations.ModastStateEnum;
import br.ufpr.inf.rp09.testdefinitions.ModastTestPlan;
import br.ufpr.inf.rp09.testdefinitions.enumerations.TestCaseEnum;
import br.ufpr.inf.rp09.testdefinitions.factories.ModastTestPlanFactory;
import br.ufpr.inf.rp09.testresults.MonitorLogResults;
import br.ufpr.inf.rp09.testresults.PerformanceInputsCalculator;
import br.ufpr.inf.rp09.testresults.TestResultAggregator;
import br.ufpr.inf.rp09.testresults.TestResultParser;
import br.ufpr.inf.rp09.testresults.TestResultsCalculator;

/**
 * Implementa o controlador de testes para testes em banco de dados VoltDB.
 * O teste e integralmente centralizado nesta classe.
 * Esta classe e responsavel por:<ul>
 * <li>configurar o plano de teste de acordo com o caso de teste escolhido,</li>
 * <li>aplicar o plano de teste aos clientes de teste local e remotos</li>
 * <li>aguardar pelo encerramento da iteracao de teste em todos os clientes</li>
 * <li>direcionar os resultados para avaliacao</li>
 * <li>verificar se houve transicao de estado na maquina de estados</li> 
 * </ul>
 * @author Rodrigo Piassetta / GRR20094994
 */
public final class VoltDBTestController extends AbstractTestController {

	/* constantes de configuracao */
	final static String RESULT_COLLECTOR_LABEL = "Summary_Test";
	final static String LOGFILE_PATH = "logs/";
	final static String LOGFILE_PREFIX = "log";
	final static String LOGFILE_EXT = ".jtl";
	final static int INITIAL_TEST_STEP = 1;
	private final static String SAMPLE_NAME_PREFIX = "TEST_STEP_";
	private final static String RESULTS_HEADER = "testStep|startTime|endTime|state|transRqts|transTreated|transRqts/s|transTreated/s|perfVar|transThroughput|perfTrend";
	private static final String RUNNING_TEST_EXCEPTION_MSG = "Nao foi possivel iniciar o plano de testes no cliente";
	final static TestCaseEnum DEFAULT_TESTCASE = TestCaseEnum.CONNECTIONS_STRESS_TESTCASE; //caso de teste padrao / stress de conexoes */

	/* atributos de controle */
	private ModastStateMachine modastStateMachine; //Maquina de estados
	private MonitorLogResults monitorLogResults; //monitora os resultados de cada iteracao de teste
	private TestResultParser testResultParser; //parser dos logs de resultados
	private TestResultAggregator testResultAggregator; //agrupa os resultados do teste	
	private StandardJMeterEngine mainTestClient;	//cliente de teste local/principal
	private Map<String, ClientJMeterEngine> mapRemoteTestClients; //clientes de teste remotos
	private TestCaseEnum testCase; //caso de teste
	private ModastTestPlan testPlan; //plano de teste
	private int testStep; //controla a iteracao do teste em execucao (base 1)

	/* valores de conexao com o banco de dados em teste */
	private String databaseAddress;
	private String databasePort;
	private String databaseName;
	private String username;
	private String password;
	
	/**
	 * Construtor da classe DBTestController
	 */
	public VoltDBTestController() {
		this(DEFAULT_TESTCASE);
	}
	
	/**
	 * Construtor principal da classe DBTestController
	 * 
	 * @param testCase Caso de teste a ser aplicado pelo controlador:
	 * <ul>
	 * <li>DBTestCaseEnum.CONNECTIONS_STRESS_TESTCASE - teste de stress sobre o numero de conexoes concorrentes suportadas pelo banco de dados</li>
	 * <li>DBTestCaseEnum.TRANSACTIONS_STRESS_TESTCASE - teste de stress sobre o numero de transacoes concorrentes suportadas pelo banco de dados</li>
	 * </ul>
	 */
	public VoltDBTestController(TestCaseEnum testCase) {
		super();
		//define caso de teste a ser aplicado
		this.testCase = testCase;
		//inicializa a lista de clientes de teste
		this.mapRemoteTestClients = new HashMap<String, ClientJMeterEngine>();
	}

	/**
	 * Cria a maquina de estados e configura para o estado inicial
	 */
	@Override
	protected void createStateMachine() {
		this.modastStateMachine = new ModastStateMachine();
		this.modastStateMachine.set2InitialState();		
	}

	/**
	 * Configura o contador de iteracoes de teste para o valor inicial
	 */
	@Override
	protected void initTestStep() {
		this.testStep = INITIAL_TEST_STEP;		
	}

	/**
	 * Cria o monitor que verifica o numero de transacoes executadas por cada 
	 * cliente de teste atraves dos logs de resultados e determina se 
	 * o teste esta em andamento ou finalizado
	 */
	@Override
	protected void createMonitorLogResults() {
		this.monitorLogResults = new MonitorLogResults();		
	}

	/**
	 * Cria o objeto respons�vel por agrupar os resultados dos testes
	 */
	@Override
	protected void createTestResultAggregator() {
		this.testResultAggregator = new TestResultAggregator();		
	}
	
	/**
	 * Configura o plano de testes de acordo com o caso de teste escolhido
	 * e a iteracao a ser executada
	 */
	@Override
	protected void configureTestPlan() {
		//cria plano de teste
		this.testPlan = ModastTestPlanFactory.createVoltDBModastTestPlan(this.getDatabaseAddress(), this.getDatabasePort(), this.getDatabaseName(), this.getUsername(), this.getPassword());
		//define arquivo de log de resultados para este plano de teste
		String logFilename = SAMPLE_NAME_PREFIX+this.testStep+"_"+System.currentTimeMillis();
		this.testPlan.getTestResultCollector().setFilename(LOGFILE_PATH+LOGFILE_PREFIX+logFilename+LOGFILE_EXT);
		
		// configura de acordo com caso de teste
		// Connection Stress: num_conn = testStep*10 transactions = 1
		// Transaction Stress: num_conn = 10 transactions = testStep*10
		// numero total dividido entre os clientes de teste
		int numClients = mapRemoteTestClients.size()+1;
		int numConn;
		switch(testCase) {
			case CONNECTIONS_STRESS_TESTCASE:
				numConn = Math.round((this.testStep*10)/numClients);
				if(numConn < 1)
					numConn = 1;
				this.testPlan.setNumThreads(numConn);
				this.testPlan.setNumLoops(1);
				this.testPlan.setSampleName(SAMPLE_NAME_PREFIX+testStep);
				break;
			case TRANSACTIONS_STRESS_TESTCASE:
				numConn = Math.round(10/numClients);
				if(numConn < 1)
					numConn = 1;
				this.testPlan.setNumThreads(numConn);
				this.testPlan.setNumLoops(this.testStep*10); 
				this.testPlan.setSampleName(SAMPLE_NAME_PREFIX+testStep);
				break;
		}
		
		//configura os clientes de teste com o plano de teste criado
		for(ClientJMeterEngine client : this.mapRemoteTestClients.values()) {
			client.configure(this.testPlan);
		}
		this.mainTestClient.configure(this.testPlan);		
	}

	/**
	 * Configura monitor do log de resultados
	 */
	@Override
	protected void configureMonitorLogResults() {
		//define caminho do arquivo de log de resultados (ja configurado no plano de testes)
		this.monitorLogResults.setLogFilePath(this.testPlan.getTestResultCollector().getFilename());
		//configura rotulo identificador das amostras a serem executadas
		this.monitorLogResults.setSampleLabel(SAMPLE_NAME_PREFIX+testStep);
		int totalClients = mapRemoteTestClients.size() + 1;
		int expectedTotalRequests = this.testPlan.getNumThreads() * this.testPlan.getNumLoops() * totalClients;
		//define numero de requisicoes a serem executadas
		this.monitorLogResults.setExpectedTotalRequests(expectedTotalRequests);
	}

	/**
	 * Executa o plano de testes nos clientes de teste 
	 * 
	 * @throws TestClientException Se nao for possivel executar o plano de testes em algum dos clientes de teste
	 */	
	@Override
	protected void runTestClients() throws TestClientException {
		//executa o plano de testes nos clientes remotos
		for(ClientJMeterEngine remoteClient : mapRemoteTestClients.values()) {
			try {
				remoteClient.runTest();
			} catch (JMeterEngineException e) {
				throw new TestClientException(RUNNING_TEST_EXCEPTION_MSG, e.getCause());
			}
		}
		//executa o plano de testes no cliente principal
		this.mainTestClient.run();		
	}

	/**
	 * Verifica se o teste esta em execucao
	 * 
	 * @return true se existe algum cliente executando o plano de testes, senao false
	 */
	@Override
	protected boolean isTestRunning() {
		//enquanto o cliente principal esta ativo nao e necessario verificar o arquivo de log
		if(this.mainTestClient.isActive())
			return true;
		
		//verifica se todas as requisicoes de teste foram registradas/executadas
		return !this.monitorLogResults.isAllRequestsLogged();
	}

	/**
	 * Processa e analisa os logs de resultados dos testes
	 */
	@Override
	protected void processTestResults() {
		this.testResultParser = new TestResultParser();
		testResultParser.processResults(this.monitorLogResults.getSampleRequesteds(), this.monitorLogResults.getSampleLabel());
	}

	/**
	 * Avalia e armazena os resultados da iteracao de teste
	 * 
	 * @param testStartTime timestamp do inicio da iteracao
	 * @param testEndTime timestamp do final da iteracao
	 */
	@Override
	protected void evaluateTestResults(long testStartTime, long testEndTime) {
		double elapsedTimeSec, transactionsRequestedPerSec,transactionsTreatedPerSec,
		performanceVariation, transactionsThroughput, performanceTrend;
		int indexResults = this.testStep-1; //(base 0)

		//converte tempo de execucao do teste de milissegundos para segundos
		elapsedTimeSec = TestResultsCalculator.convertElapsedTimeMillis2Seconds(testEndTime - testStartTime);
		
		//calcula taxa de transacoes requisitadas e tratadas por segundo
		transactionsTreatedPerSec = TestResultsCalculator.calculateRatePerSecond(this.testResultParser.getSuccessfulRequests(), elapsedTimeSec);
		transactionsRequestedPerSec = TestResultsCalculator.calculateRatePerSecond(this.testResultParser.getTotalRequests(), elapsedTimeSec);

		//adiciona resultados (indexado por testStep-1 / 0-based)
		this.testResultAggregator.setTestStartTime(indexResults, testStartTime);
		this.testResultAggregator.setTestEndTime(indexResults, testEndTime);
		this.testResultAggregator.setTestState(indexResults, this.modastStateMachine.getState());
		this.testResultAggregator.setTransactionsRequested(indexResults, this.testResultParser.getTotalRequests());
		this.testResultAggregator.setTransactionsTreated(indexResults, this.testResultParser.getSuccessfulRequests());
		this.testResultAggregator.setTransactionsRequestedRate(indexResults, transactionsRequestedPerSec);
		this.testResultAggregator.setTransactionsTreatedRate(indexResults, transactionsTreatedPerSec);
		
		//calcula delta/teta/gama
		performanceVariation = PerformanceInputsCalculator.calculatePerformanceVariation(this.testStep, this.testResultAggregator.getAllTransactionsTreatedRate());
		transactionsThroughput = PerformanceInputsCalculator.calculateTransactionTroughput(transactionsTreatedPerSec, transactionsRequestedPerSec);
		performanceTrend = PerformanceInputsCalculator.calculatePerformanceTrend();
		
		//adiciona resultados (indexado por testStep-1 / 0-based)
		this.testResultAggregator.setPerformanceVariation(indexResults, performanceVariation);
		this.testResultAggregator.setTransactionsThroughput(indexResults, transactionsThroughput);
		this.testResultAggregator.setPerformanceTrend(indexResults, performanceTrend);
	}

	/**
	 * Verifica se houve transicao de estado de acordo com os resultados obtidos
	 * e configura o novo estado na maquina de estados
	 */	
	@Override
	protected void checkStateTransition() {
		int indexResults = this.testStep-1;
		//avalia qual deve ser o proximo estado de acordo com os resultados do teste
		ModastStateEnum nextState = ModastStateEvaluator.getNextState(this.modastStateMachine.getState(), 
															  this.testResultAggregator.getPerformanceVariation(indexResults), 
															  this.testResultAggregator.getTransactionsThroughput(indexResults), 
															  this.testResultAggregator.getPerformanceTrend(indexResults),
															  this.testResultAggregator.getTransactionsTreated(indexResults));
		
		//se ha transicao de estado, configura a maquina de estados para o novo estado
		if(nextState != modastStateMachine.getState())
			this.modastStateMachine.setStateTransition(nextState, this.testPlan.getTestResultCollector().getEndTime());		
	}

	/**
	 * Incrementa o contador de iteracoes de teste
	 */
	@Override
	protected void nextTestStep() {
		this.testStep++;	
	}
	
	/**
	 * Define o cliente de testes principal
	 * @param testClient Cliente JMeter rodando na mesma maquina que a aplicacao de testes
	 */
	public void setMainTestClient(StandardJMeterEngine testClient) {
		this.mainTestClient = testClient;		
	}

	/**
	 * Adiciona os clientes de teste remotos que serao utilizados nos testes
	 * 
	 * @param hostName Hostname ou endereco IP do cliente adicionado
	 * @param testClient Cliente JMeter rodando em maquina remota
	 */
	public void addRemoteTestClient(String hostName, ClientJMeterEngine testClient) {
		this.mapRemoteTestClients.put(hostName, testClient);		
	}

	/**
	 * Imprime resultado final dos testes na saida indicada por parametro
	 * 
	 * @param outputResults arquivo de saida para impressao dos resultados
	 */
	public void printResults(PrintStream outputResults) {
		if(this.testStep-1 == INITIAL_TEST_STEP) {
			outputResults.println(RESULTS_HEADER);
		}
		String strResult;
		int i = this.testStep - 2;
		strResult = String.valueOf(i+1);
		strResult = strResult.concat("|"+this.testResultAggregator.getTestStartTime(i));
		strResult = strResult.concat("|"+this.testResultAggregator.getTestEndTime(i));
		strResult = strResult.concat("|"+this.testResultAggregator.getTestState(i));
		strResult = strResult.concat("|"+this.testResultAggregator.getTransactionsRequested(i));
		strResult = strResult.concat("|"+this.testResultAggregator.getTransactionsTreated(i));
		strResult = strResult.concat("|"+Double.toString(this.testResultAggregator.getTransactionsRequestedRate(i)));
		strResult = strResult.concat("|"+Double.toString(this.testResultAggregator.getTransactionsTreatedRate(i)));
		strResult = strResult.concat("|"+Double.toString(this.testResultAggregator.getPerformanceVariation(i)));
		strResult = strResult.concat("|"+Double.toString(this.testResultAggregator.getTransactionsThroughput(i)));
		strResult = strResult.concat("|"+Double.toString(this.testResultAggregator.getPerformanceTrend(i)));
			
		outputResults.println(strResult);
	}
	 
	public String getDatabaseAddress() {
		return databaseAddress;
	}

	public void setDatabaseAddress(String databaseAddress) {
		this.databaseAddress = databaseAddress;
	}

	public String getDatabasePort() {
		return databasePort;
	}

	public void setDatabasePort(String databasePort) {
		this.databasePort = databasePort;
	}

	public String getDatabaseName() {
		return databaseName;
	}

	public void setDatabaseName(String databaseName) {
		this.databaseName = databaseName;
	}

	public String getUsername() {
		return username;
	}

	public void setUsername(String username) {
		this.username = username;
	}

	public String getPassword() {
		return password;
	}

	public void setPassword(String password) {
		this.password = password;
	}

	/**
	 * Remove os arquivos de log de resultados da pasta de trabalho
	 * e os arquiva em uma pasta dedicada se saveLogFiles for true
	 * 
	 * @param saveLogFiles arquiva os arquivos de log em uma pasta dedicada se true,
	 * caso contrario apenas os remove da pasta de trabalho
	 */
	public void removeLogFiles(boolean saveLogFiles) {
		File logFolder = new File(LOGFILE_PATH);
		File newFolder = null;
		File[] logFiles = logFolder.listFiles();
		if(saveLogFiles) {
			newFolder = new File(LOGFILE_PATH+"logs"+System.currentTimeMillis());
			newFolder.mkdir();
		}
		
		for(File logFile : logFiles) {
		    if(logFile.getName().endsWith(LOGFILE_EXT)) {
		    	if(saveLogFiles)
		    		logFile.renameTo(new File(newFolder, logFile.getName()));
		    	else
		    		logFile.delete();
		    }
		}		
	}

	public ModastStateMachine getModastStateMachine() {
		return this.modastStateMachine;
	}

}
