package br.ufpr.inf.rp09.statemachine;

import br.ufpr.inf.rp09.statemachine.enumerations.ModastStateEnum;

/**
 * Avalia o estado do banco de dados a partir de variaveis de desempenho
 * 
 * @author Rodrigo Piassetta / GRR20094994
 *
 */
public class ModastStateEvaluator {

	final static double WARMUP_THRESHOLD = 0.1;
	final static double STEADY_THRESHOLD = 0.9;
	final static double STRESS_THRESHOLD = 0.1;
	final static double THRASHING_THRESHOLD = 1.0;
	
	/**
	 * Verifica o estado no qual o banco de dados deve se encontrar
	 * de acordo com as variaveis de desempenho e seu estado atual
	 * 
	 * @param currentState
	 * @param performanceVariation
	 * @param transactionThroughput
	 * @param performanceTrend
	 * 
	 * @return estado do banco de dados definido pela analise das variaveis de desempenho
	 */
	public static ModastStateEnum getNextState(ModastStateEnum currentState, double performanceVariation, double transactionThroughput,
			double performanceTrend, double transactionsTreated) {
		ModastStateEnum nextState;
		
		switch(currentState) {
			case WARMUP_STATE:
				nextState = warmUpStateDecision(performanceVariation);
				break;
			case STEADY_STATE:
				nextState = steadyStateDecision(transactionThroughput);
				break;
			case UNDER_PRESSURE_STATE:
				nextState = underPressureStateDecision(performanceVariation, transactionThroughput, transactionsTreated);
				break;
			case STRESS_STATE:
				nextState = stressStateDecision(performanceVariation, transactionThroughput, performanceTrend);
				break;
			case THRASHING_STATE:
			default:
				nextState = currentState;
				break;
		}
		
		return nextState;
	}

	/**
	 * Verifica se ha mudanca de estado quando o banco de dados se encontra em WarmUp State
	 * 
	 * @param performanceVariation
	 * 
	 * @return O novo estado do banco de dados definido pela analise das variaveis 
	 */
	private static ModastStateEnum warmUpStateDecision(double performanceVariation) {
		return (performanceVariation < WARMUP_THRESHOLD)?ModastStateEnum.STEADY_STATE:ModastStateEnum.WARMUP_STATE;
	}

	/**
	 * Verifica se ha mudanca de estado quando o banco de dados se encontra em Steady State
	 * 
	 * @param transactionThroughput
	 * 
	 * @return O novo estado do banco de dados definido pela analise das variaveis
	 */
	private static ModastStateEnum steadyStateDecision(double transactionThroughput) {
		return (transactionThroughput > STEADY_THRESHOLD)?ModastStateEnum.STEADY_STATE:ModastStateEnum.UNDER_PRESSURE_STATE;
	}

	/**
	 * Verifica se ha mudanca de estado quando o banco de dados se encontra em Under Pressure State
	 * 
	 * @param performanceVariation
	 * @param transactionThroughput
	 * 
	 * @return O novo estado do banco de dados definido pela analise das variaveis
	 */
	private static ModastStateEnum underPressureStateDecision(double performanceVariation, double transactionThroughput, double transactionsTreated) {
		if(performanceVariation > STRESS_THRESHOLD*transactionsTreated) 
			return ModastStateEnum.STRESS_STATE;
		
		if(transactionThroughput > STEADY_THRESHOLD)
			return ModastStateEnum.STEADY_STATE;
		
		return ModastStateEnum.UNDER_PRESSURE_STATE;
	}
	
	/**
	 * Verifica se ha mudanca de estado quando o banco de dados se encontra em Stress State
	 * 
	 * @param performanceVariation
	 * @param transactionThroughput
	 * @param performanceTrend
	 * 
	 * @return O novo estado do banco de dados definido pela analise das variaveis
	 */
	private static ModastStateEnum stressStateDecision(double performanceVariation, double transactionThroughput, double performanceTrend) {
		if(performanceTrend < THRASHING_THRESHOLD*transactionThroughput)
			return ModastStateEnum.THRASHING_STATE;
		
		if(performanceVariation <= STRESS_THRESHOLD)
			return ModastStateEnum.UNDER_PRESSURE_STATE;
		
		return ModastStateEnum.STRESS_STATE;
	}
}
