package br.ufpr.inf.rp09.testdefinitions.factories;

import org.apache.jmeter.config.RandomVariableConfig;
import org.apache.jmeter.control.LoopController;
import org.apache.jmeter.protocol.jdbc.config.DataSourceElement;
import org.apache.jmeter.protocol.jdbc.sampler.JDBCSampler;
import org.apache.jmeter.reporters.Summariser;
import org.apache.jmeter.testelement.TestPlan;
import org.apache.jmeter.threads.ThreadGroup;
import org.apache.jorphan.collections.HashTree;
import br.ufpr.inf.rp09.testdefinitions.ModastTestPlan;
import br.ufpr.inf.rp09.testresults.TestResultCollector;

/**
 * Fabrica de planos de teste para JMeter
 * 
 * @author Rodrigo Piassetta / GRR20094994
 *
 */
public class ModastTestPlanFactory {

	private static final String POSTGRESQL_URL_PREFIX = "jdbc:postgresql://";
	private static final String POSTGRESQL_JDBC_DRIVER = "org.postgresql.Driver";
	private static final String VOLTDB_URL_PREFIX = "jdbc:voltdb://";
	private static final String VOLTDB_JDBC_DRIVER = "org.voltdb.jdbc.Driver";
	
	/**
	 * Cria um plano de teste base para banco de dados VoltDB
	 * 
	 * @param dbAddress endereco IP do banco de dados
	 * @param dbPort porta do banco de dados
	 * @param databaseName nome da base de dados
	 * @param username nome de usuario
	 * @param password senha do usuario
	 * 
	 * @return O plano de teste base para banco de dados VoltDB
	 */	
	public static ModastTestPlan createVoltDBModastTestPlan(String dbAddress, String dbPort, String databaseName, String username, String password) {
		
		ModastTestPlan testPlanTree = new ModastTestPlan();
		
        // Loop Controller
        LoopController loopController = createLoopController(); 

        // Thread Group
        ThreadGroup threadGroup = createThreadGroup();
        threadGroup.setSamplerController(loopController);
		
        // JDBC Data Source
        String dbUrl = VOLTDB_URL_PREFIX+dbAddress+":"+dbPort+"/"+databaseName;
        //String dbUrl = VOLTDB_URL_PREFIX+dbAddress+":"+dbPort;
        DataSourceElement dataSource = createJDBCDataSourceElement();
        dataSource.setProperty("dbUrl", dbUrl);
        dataSource.setProperty("driver", VOLTDB_JDBC_DRIVER);
        dataSource.setProperty("username", username);
        dataSource.setProperty("password", password);

        // JDBC Sampler Request
        JDBCSampler jdbcSampler = createJDBCSampler();
		jdbcSampler.setProperty("queryType", "Callable Statement");
		jdbcSampler.setProperty("query", "BEGIN TRANSACTION;"
									   + "UPDATE accounts SET Abalance = Abalance + ? WHERE Aid = ?;"
									   + "SELECT Abalance FROM accounts WHERE Aid = ?;"
									   + "UPDATE tellers SET Tbalance = Tbalance + ? WHERE Tid = ?;"
									   + "UPDATE branches SET Bbalance = Bbalance + ? WHERE Bid = 1;"
									   + "INSERT INTO history(Tid, Bid, Aid, delta, time) VALUES (?, 1, ?, ?, CURRENT_TIMESTAMP);"
									   + "COMMIT TRANSACTION;");
		jdbcSampler.setProperty("queryArguments", "${rndDelta},${rndAid},${rndAid},${rndDelta},${rndTid},${rndDelta},${rndTid},${rndAid},${rndDelta}");
		jdbcSampler.setProperty("queryArgumentsTypes", "INTEGER,INTEGER,INTEGER,INTEGER,INTEGER,INTEGER,INTEGER,INTEGER,INTEGER");

		RandomVariableConfig randomDelta = new RandomVariableConfig();
		randomDelta.setProperty("minimumValue", "-999999");
		randomDelta.setProperty("maximumValue", "999999");
		randomDelta.setProperty("perThread", false);
		randomDelta.setProperty("variableName", "rndDelta");
		
		RandomVariableConfig randomAid = new RandomVariableConfig();
		randomAid.setProperty("minimumValue", "1");
		randomAid.setProperty("maximumValue", "100000");
		randomAid.setProperty("perThread", false);
		randomAid.setProperty("variableName", "rndAid");

		RandomVariableConfig randomTid = new RandomVariableConfig();
		randomTid.setProperty("minimumValue", "1");
		randomTid.setProperty("maximumValue", "10");
		randomTid.setProperty("perThread", false);
		randomTid.setProperty("variableName", "rndTid");
		
		TestPlan testPlan = new TestPlan();
		testPlanTree.add(testPlan);
		HashTree threadGroupTree = testPlanTree.add(testPlan, threadGroup);
		threadGroupTree.add(randomDelta);
		threadGroupTree.add(randomAid);
		threadGroupTree.add(randomTid);
		threadGroupTree.add(jdbcSampler);
		threadGroupTree.add(dataSource);

        // Result Collector
        TestResultCollector resultCollector = new TestResultCollector(new Summariser("Summary_VoltDB_Test"));
        testPlanTree.add(testPlanTree.getArray()[0], resultCollector);		

        // Setting control elements
        testPlanTree.setThreadGroup(threadGroup);
        testPlanTree.setLoopController(loopController);
        testPlanTree.setSampler(jdbcSampler);
        testPlanTree.setTestResultCollector(resultCollector);
		
		return testPlanTree;
	}
	
	/**
	 * Cria um plano de teste base para banco de dados PostgreSQL
	 * 
	 * @param dbAddress endereco IP do banco de dados
	 * @param dbPort porta do banco de dados
	 * @param databaseName nome da base de dados
	 * @param username nome de usuario
	 * @param password senha do usuario
	 * 
	 * @return O plano de teste base para banco de dados PostgreSQL
	 */
	public static ModastTestPlan createPostgreSQLModastTestPlan(String dbAddress, String dbPort, String databaseName, String username, String password) {
		
		ModastTestPlan testPlanTree = new ModastTestPlan();
		
        // Loop Controller
        LoopController loopController = createLoopController(); 

        // Thread Group
        ThreadGroup threadGroup = createThreadGroup();
        threadGroup.setSamplerController(loopController);
		
        // JDBC Data Source
        String dbUrl = POSTGRESQL_URL_PREFIX+dbAddress+":"+dbPort+"/"+databaseName;
        DataSourceElement dataSource = createJDBCDataSourceElement();
        dataSource.setProperty("dbUrl", dbUrl);
        dataSource.setProperty("driver", POSTGRESQL_JDBC_DRIVER);
        dataSource.setProperty("username", username);
        dataSource.setProperty("password", password);

        // JDBC Sampler Request
        JDBCSampler jdbcSampler = createJDBCSampler();
		jdbcSampler.setProperty("queryType", "Callable Statement");
		jdbcSampler.setProperty("query", "BEGIN TRANSACTION;"
									   + "UPDATE accounts SET Abalance = Abalance + ? WHERE Aid = ?;"
									   + "SELECT Abalance FROM accounts WHERE Aid = ?;"
									   + "UPDATE tellers SET Tbalance = Tbalance + ? WHERE Tid = ?;"
									   + "UPDATE branches SET Bbalance = Bbalance + ? WHERE Bid = 1;"
									   + "INSERT INTO history(Tid, Bid, Aid, delta, time) VALUES (?, 1, ?, ?, CURRENT_TIMESTAMP);"
									   + "COMMIT TRANSACTION;");
		jdbcSampler.setProperty("queryArguments", "${rndDelta},${rndAid},${rndAid},${rndDelta},${rndTid},${rndDelta},${rndTid},${rndAid},${rndDelta}");
		jdbcSampler.setProperty("queryArgumentsTypes", "INTEGER,INTEGER,INTEGER,INTEGER,INTEGER,INTEGER,INTEGER,INTEGER,INTEGER");

		RandomVariableConfig randomDelta = new RandomVariableConfig();
		randomDelta.setProperty("minimumValue", "-999999");
		randomDelta.setProperty("maximumValue", "999999");
		randomDelta.setProperty("perThread", false);
		randomDelta.setProperty("variableName", "rndDelta");
		
		RandomVariableConfig randomAid = new RandomVariableConfig();
		randomAid.setProperty("minimumValue", "1");
		randomAid.setProperty("maximumValue", "100000");
		randomAid.setProperty("perThread", false);
		randomAid.setProperty("variableName", "rndAid");

		RandomVariableConfig randomTid = new RandomVariableConfig();
		randomTid.setProperty("minimumValue", "1");
		randomTid.setProperty("maximumValue", "10");
		randomTid.setProperty("perThread", false);
		randomTid.setProperty("variableName", "rndTid");
		
		TestPlan testPlan = new TestPlan();
		testPlanTree.add(testPlan);
		HashTree threadGroupTree = testPlanTree.add(testPlan, threadGroup);
		threadGroupTree.add(randomDelta);
		threadGroupTree.add(randomAid);
		threadGroupTree.add(randomTid);
		threadGroupTree.add(jdbcSampler);
		threadGroupTree.add(dataSource);

        // Result Collector
        TestResultCollector resultCollector = new TestResultCollector(new Summariser("Summary_PostgreSQL_Test"));
        testPlanTree.add(testPlanTree.getArray()[0], resultCollector);		

        // Setting control elements
        testPlanTree.setThreadGroup(threadGroup);
        testPlanTree.setLoopController(loopController);
        testPlanTree.setSampler(jdbcSampler);
        testPlanTree.setTestResultCollector(resultCollector);
		
		return testPlanTree;
	}
	
	private static LoopController createLoopController() {
		LoopController loopController = new LoopController();
		loopController.setLoops(0);
		loopController.setFirst(true);
		loopController.setContinueForever(false);
		loopController.initialize();

		return loopController;
	}

	private static ThreadGroup createThreadGroup() {
        ThreadGroup threadGroup = new ThreadGroup();
        threadGroup.setNumThreads(0);
        threadGroup.setRampUp(0);
        threadGroup.setScheduler(false);

		return threadGroup;
	}

	private static DataSourceElement createJDBCDataSourceElement() {
        DataSourceElement dataSource = new DataSourceElement();
        dataSource.setProperty("dataSource", "jdbcpool");
        dataSource.setProperty("poolMax", "10");
        dataSource.setProperty("timeout", "10000");
        dataSource.setProperty("trimInterval", "60000");
        dataSource.setProperty("autocommit", false);
        dataSource.setProperty("transactionIsolation", "DEFAULT");
        dataSource.setProperty("keepAlive", true);
        dataSource.setProperty("connectionAge", "5000");
        dataSource.setProperty("checkQuery", "Select 1");

		return dataSource;
	}
	
	private static JDBCSampler createJDBCSampler() {
		JDBCSampler jdbcSampler = new JDBCSampler();
		jdbcSampler.setProperty("dataSource", "jdbcpool");

		return jdbcSampler;
	}	
}
